! pi_bench.f - calculate pi as a long string of digits using
!              John Machin's approach.
!              Created for Fortran-66: August  14 1989.
!              Updated for Fortran-77: April   29 1992.
!              Updated for GFortran:   October 26 2017.
!              Last update: 09/12/2021 (dd/mm/yyyy)
! (C) 1989--2021 AIAP by Golub M.A. Moscow Russia mike-golub@yandex.ru
! ======================================================================

Program pi_bench

Parameter (NDM=64000, NDM1=NDM+1) ! NB: In.n and low limit
Integer  NB(NDM1), NC(NDM1), N1(NDM1), N2(NDM1)

Character * 10 stm / 'Start time' /, etm / 'End time  ' /
Character * 30 tm, tm1
Character * 50 name, dummy

Common / PIDIM0 / IDM0, NSQRU
Common / PIUNIT / NUNIT
Data   ndim / NDM /, N5 / 5 /, N239 / 239 /, N4 / 4 /

NUNIT=10000

! NUNIT -- program tuning number (fortran dependent).
! This number must be less power than REAL type significant digits number.

RUNIT=NUNIT; NSQRU=Int(Sqrt(RUNIT)); NDIGIT=Int(ALog10(RUNIT)+0.5); ndmm=NDIGIT * NDM

Print "('Pi_bench -- computing power benchmark based on Machin''s pi number computing '/'algorithm.'//" // &
"'If computing duration is more than 1 second then the program updates pi_bench.dat file:'/" // &
"'the last line of the file must be a 50 characters (at most) long name of your computing'/" // &
"'environment,  e.g., ''Cygwin gfortran 64 bit @ i5-7500 3.4 GHz 16 GB RAM''.'/" // &
"'The line will be copied and appended with its benchmark value.'/" // &
"'Two else files are created/updated: ''pi.tst'' and ''pi.out''. The first one'/" // &
"'contains timimg data, the second contains computed values of arctangents of 1/5'/" // &
"'and 1/239 and the pi value.'//'Hit Ctrl+D to quit.'/)"

Do
   Print "('Enter number of digits for pi [4;',i6,'] ',$)", ndmm
   Read(*, *, IOStat=io) ndim;
   If (io/=0) Then; Print *, 'Invalid value specified.'; Call Exit(901); End If
   Print *, ndim, ' has been entered.'
   If (ndim<4) Cycle
   If (ndim<=ndmm) Exit
   Print "('You misspelled.',i7,' is maximum. Retry, please.')", ndmm
End Do

IDM0=(ndim-1)/NDIGIT+1; IDM1=IDM0*NDIGIT; ndim=IDM0+1

Print *, IDM1, ' has been accepted.'

Open(1, File='pi.tst', IOStat=io)
If (io/=0) Then; Print *, 'Error opening pi.tst'; Stop 1000; End If

Open(3, File='pi.out', IOStat=io)
If (io/=0) Then; Print *, 'Error opening pi.out'; Stop 1001; End If

! Call Time(tm)
Call FDate(tm)

Print 1803, ndim
Write(1, 1803, Err=903) ndim
1803 Format('Speed test data: dimension is ', i5)
Print 1804, stm, tm(12:19) ! hh, mm, ss
Write(1, 1804, Err=903) stm, tm(12:19) ! hh, mm, ss
1804 Format(a, 1x, a) ! 2(i2.2, ':'), i2.2, '.', i2.2)

Print *, 'Calculating arctg(1/5).'
ds=Secnds(0.)
Call ArcTg(N1, NB, NC, ndim, N5)

! Print 1805, (N1(I), I=1, IDM0)
! 1805 Format(' ArcTangent(1 / 5)=0.', (T26, 13 I4.4))
Write(3, 1801, Err=903) (N1(I), I=1, IDM0)
1801 Format('ArcTangent(1 / 5)=0.', (T21, 25 I4.4))

Print *, 'Calculating arctg(1/239).'
Call ArcTg(N2, NB, NC, ndim, N239)

! Print 1806, (N2(I), I=1, IDM0)
! 1806 Format(' ArcTangent(1/239)=0.', (T26, 13 I4.4))
Write(3, 1802, Err=903) (N2(I), I=1, IDM0)
1802 Format(/'ArcTangent(1/239)=0.',(T21, 25I4.4))

Print *, 'Computing arctg(1/5)*4.'
Call Mult (N1, ndim, N4)
Print *, 'Computing arctg(1/5)*4-arctg(1/239).'
Call Subtr(N1, N2, ndim)
Print *, 'Computing pi=(arctg(1/5)*4-arctg(1/239))*4.'
Call Mult (N1, ndim, N4)

N1(1)=Mod(N1(1), NUNIT)

ds=Secnds(ds)
! Call Time(tm1)
Call FDate(tm1)

Print 1804, etm, tm1(12:19) ! h1, m1, s1
Write(1, 1804, Err=903) etm, tm1(12:19) ! h1, m1, s1
Print 10, ds
Write(1, 10, Err=903) ds
10 Format('Duration is ',g16.8,' seconds')
! Print 1808, (N1(I), I=1, IDM0)
! 1808 Format(' Pi=3.', (T9, 15 I4.4))
Write(3, 1800, Err=903) (N1(I), I=1, IDM0)
1800 Format( /'Pi=3.',(T6,25I4.4))

rdm=idm1

If (ds>=1.) Then
   bench=(rdm*6e-4*rdm)/ds
   Open(2, File='pi_bench.dat', Status='Old', IOStat=io)
   If (io/=0) Then; Print *, 'Error opening pi_bench.dat for reading'; call Exit(1002); End If

   Do; Read(2, "(a50)", Err=912, End=911) dummy; name=dummy; End Do

   Print *, 'Computing environment name="', name, '"'

   911 Continue

   Close(2, IOStat=io); If (io/=0) Then; Print *, 'Error closing pi_bench.dat'; Call Exit(1003); End If

   Open(2, File='pi_bench.dat', Status='Unknown', Position='Append', IOStat=io)
   If (io/=0) Then; Print *, 'Error opening pi_bench.dat for writing'; Call Exit(1004); End If
   Print "(/'Benchmark value for ',a,' is ',g15.7)", name(1:len_trim(name)), bench
   Write(2, "(a50,1x,g15.7)", Err=912) name, bench
Else
   Print "('Computation duration is less than 1 second: pi_bench.dat has not been updated.')"
End If

Call Exit(0)
903 Continue; Print *, 'Write error.'; Call Exit(903)
912 Continue; Print *, 'Error reading/writing benchmark file.'; Call Exit(912)
End Program

! Arctangent calculation for 1 / N argument. Long digits string NA(ndim) is formed.

Subroutine ArcTg(NA, NB, NC, ndim, N)

    Integer   NA(ndim), NB(ndim), NC(ndim)

    External  FN

    Common / PIDIM0 / IDM0, NSQRU
    Common / PIUNIT / NUNIT
    Common / PIFUNC / Alpha, Beta

    If (N<NSQRU) Then; N2=N*N; I2=1; Else; N2=N; I2=2; End If

    NC(1)=NUNIT
    Call Zero(NC, ndim, 2)

    Beta=1./ALog(1.*N); NDIM1=Int(ALog10(1.*NUNIT)+0.5)*ndim
    !              ---------- ln(10.) --------------
    Alpha=NDIM1 * 2.302585 ! 092994045684017991454684

    X=Alpha; X=RIter(FN, X, 0.5); NL=Int(X)+1

    Print *, 'Expansion length is considered:', NL
    Print *

    Call Divide(NC, ndim, N)
    Call Copy  (NC, NA, ndim)

    Do I10=2, NL, 2
       Do I20=1, I2; Call Divide(NC, ndim, N2); End Do
       J=2*I10-1
       Call Copy  (NC, NB, ndim)
       Call Divide(NB, ndim, J)
       Call Subtr (NA, NB, ndim)
       DO I30=1, I2; Call Divide(NC, ndim, N2); End Do
       Call Copy  (NC, NB, ndim)
       Call Divide(NB, ndim, J+2)
       Call Add   (NA, NB, ndim)
       ! Print "('+Term #', i5, ' calculated.')", I10
    End Do
    ! Print "('+Last operand trail: ', 2i4.4)", NB(IDM0), NB(ndim)
    Return
End Subroutine

! Function for expansion truncation iteration search.

Real Function FN(X); Common / PIFUNC / A, B; FN=((A-ALog(X))*B+1.)*0.5; Return; End Function

! Arithmetic operations package

! Assignment operation: NX -> NY
Subroutine Copy(NX, NY, ndim)
    Integer NX(ndim), NY(ndim)
    Do I10=1, ndim; NY(I10)=NX(I10); End Do
    Return
End Subroutine

! Assignment: NX=0 (all digits from J-th)
Subroutine Zero(NX, ndim, J)
    Integer NX(ndim)
    Do I10=J, ndim; NX(I10)=0; End Do
    Return
End Subroutine

! Multiplication: IARR(ndim) * MULT2 -> IARR(ndim). NB: multiplier "MULT2" is a common integer.
Subroutine Mult(IARR, ndim, MULT2)
    Integer  IARR(ndim)
    Common / PIUNIT / NUNIT
    NBUFF=0
    Do I10=1, ndim
       J=ndim-I10+1; Buffer=IARR(J); Buffer=Buffer*MULT2+NBUFF; NBUFF=Int(Buffer/NUNIT)
       IARR(J)=Int(Buffer-1.*NUNIT*NBUFF+0.5)
    End Do
    Return
End Subroutine

! Addition: IARR1+IARR2 -> IARR1
Subroutine Add(IARR1, IARR2, ndim)
    Integer IARR1(ndim), IARR2(ndim)
    Common / PIUNIT / NUNIT
    NBUFF=0
    Do I10=1, ndim
       J=ndim-I10+1; Buffer=IARR1(J); Buffer=IARR2(J)+Buffer+NBUFF; NBUFF=Int(Buffer/NUNIT)
       IARR1(J)=Int(Buffer-1.*NUNIT*NBUFF+0.5)
    End Do
    Return
End Subroutine

! Subtraction: IARR1-IARR2 -> IARR1
Subroutine Subtr(IARR1, IARR2, ndim)
    Integer IARR1(ndim), IARR2(ndim)
    Common / PIUNIT / NUNIT
    NBUFF=0
    Do I10=1, ndim
       J=ndim-I10+1; NSUBT=IARR1(J)-IARR2(J)-NBUFF; NBUFF=0
       If (NSUBT<0) Then; NSUBT=NUNIT+NSUBT; NBUFF=1; End If
       IARR1(J)=NSUBT
    End Do
    Return
End Subroutine

! Division: NARR(ndim) / NUMER -> NARR(ndim). NB: numerator is a common integer.

Subroutine Divide(NARR, ndim, NUMER)
    Integer NARR(ndim)
    Common / PIUNIT / NUNIT
    NREST=0
    Do I10=1, ndim
       Buffer=NARR(I10)+1.*NREST*NUNIT; NDIV=Int(Buffer/NUMER)
       NREST=Int(Buffer-1.*NUMER*NDIV+0.5); NARR(I10)=NDIV
    End Do
    Return
End Subroutine

! RITER - iterational procedure for x=F(x) root search.
!
! X0 - initial x approximation;
!
! Eps - some precision for the root: if previous and current estimations are closer than EPS*|Xold-Xnew|
! then the search is considered done. Is is good for the functions with some bad behavior in the sense
! of convergence.
!
! Parameter Max_Away is a permitted number of consecutive movements of a point, which represents x-value
! estimation, away form the previous one in an increasing distance.
!
! Error=0 - OK; 2001 - no convergence.
!
! (C) by M.Golub * Moscow * Russia * 1995

Real Function RITER (Funct, X0, Eps)
    Parameter (Max_Away=5)
    Integer Error
    Common / FNERR / Error

    Error=0; RITER=X0; N=0; X_old=RITER; RITER=Funct(RITER); Dist=Abs(RITER-X_old)

    Do
       Dst=Dist; X_old=RITER; RITER=Funct(RITER); Dist =Abs(RITER-X_old)
       If (Dist<=Eps*Abs(RITER)) Return
       If (Dist>=Dst) N=N+1
       If (N>Max_Away) Exit
    End Do

    Error=2001

    Return
End Function
